/////////////////////////////////////////////////////////////////////////////
// Gb3d.h : Gamebios 3D
/////////////////////////////////////////////////////////////////////////////
#ifndef __GB3D_H__
#define __GB3D_H__
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// Directive

#include "Gamebios.h"
#include "GbVideo.h"
#include "Gb2d.h"
#include "Gb3dStruct.h"

/////////////////////////////////////////////////////////////////////////////
// Prototype

#define GB3DRGB(r, g, b)			D3DRGB(r, g, b)
#define GB3DRGBA(r, g, b, a)		D3DRGBA(r, g, b, a)
#define GB3DVALUE					D3DVALUE
#define GB3DVECTOR					D3DVECTOR

#define GB3DSTARTUPDEVICE_NOZBUFFER	(0x0001)

#define GB3DOBJECT_ALPHABLEND		(0x0001)
#define GB3DOBJECT_ZBUFFER			(0x0002)

#define GB3DEFFECT_ENVIRONMENT		(1 << 1)
#define GB3DEFFECT_SPECULAR			(1 << 2)

#define GB3DMODEL_FLAT				(0x0000)
#define GB3DMODEL_GOURAUD			(0x0001)

#define GB3DVIEWPORT_DEFAULT		(0)

#define GB3DLIGHT_POINT				D3DLIGHT_POINT
#define GB3DLIGHT_SPOT				D3DLIGHT_SPOT
#define GB3DLIGHT_DIRECTIONAL		D3DLIGHT_DIRECTIONAL
#define	GB3DLIGHT_PARALLELPOINT		D3DLIGHT_PARALLELPOINT

#define GB3DLIGHT_ACTIVE			D3DLIGHT_ACTIVE
#define GB3DLIGHT_NO_SPECULAR		D3DLIGHT_NO_SPECULAR

#define GB3DLIGHT_RANGE_MAX			D3DLIGHT_RANGE_MAX

#define GB3DBLEND					D3DBLEND
#define GB3DBLEND_ZERO				D3DBLEND_ZERO
#define GB3DBLEND_ONE				D3DBLEND_ONE
#define GB3DBLEND_SRCCOLOR			D3DBLEND_SRCCOLOR
#define GB3DBLEND_INVSRCCOLOR		D3DBLEND_INVSRCCOLOR
#define GB3DBLEND_SRCALPHA			D3DBLEND_SRCALPHA
#define GB3DBLEND_INVSRCALPHA		D3DBLEND_INVSRCALPHA
#define GB3DBLEND_DSTALPHA			D3DBLEND_DESTALPHA
#define GB3DBLEND_INVDSTALPHA		D3DBLEND_INVDESTALPHA
#define GB3DBLEND_DSTCOLOR			D3DBLEND_DESTCOLOR
#define GB3DBLEND_INVDSTCOLOR		D3DBLEND_INVDESTCOLOR
#define GB3DBLEND_SRCALPHASAT		D3DBLEND_SRCALPHASAT
#define GB3DBLENDMODE				D3DTEXTUREBLEND
#define GB3DBLENDMODE_DECAL			D3DTBLEND_DECAL
#define GB3DBLENDMODE_MODULATE		D3DTBLEND_MODULATE
#define GB3DBLENDMODE_DECALALPHA	D3DTBLEND_DECALALPHA
#define GB3DBLENDMODE_MODULATEALPHA	D3DTBLEND_MODULATEALPHA
#define GB3DBLENDMODE_DECALMASK		D3DTBLEND_DECALMASK
#define GB3DBLENDMODE_MODULATEMASK	D3DTBLEND_MODULATEMASK
#define GB3DBLENDMODE_COPY			D3DTBLEND_COPY
#define GB3DBLENDMODE_ADD			D3DTBLEND_ADD

typedef enum {
	GB3DTOP_DISABLE						= D3DTOP_DISABLE,
	GB3DTOP_SELECTARG1					= D3DTOP_SELECTARG1,
	GB3DTOP_SELECTARG2					= D3DTOP_SELECTARG2,
	GB3DTOP_MODULATE					= D3DTOP_MODULATE,
	GB3DTOP_MODULATE2X					= D3DTOP_MODULATE2X,
	GB3DTOP_MODULATE4X					= D3DTOP_MODULATE4X,
	GB3DTOP_ADD							= D3DTOP_ADD,
	GB3DTOP_ADDSIGNED					= D3DTOP_ADDSIGNED,
	GB3DTOP_ADDSIGNED2X					= D3DTOP_ADDSIGNED2X,
	GB3DTOP_SUBTRACT					= D3DTOP_SUBTRACT,
	GB3DTOP_ADDSMOOTH					= D3DTOP_ADDSMOOTH,
	GB3DTOP_BLENDDIFFUSEALPHA			= D3DTOP_BLENDDIFFUSEALPHA,
	GB3DTOP_BLENDTEXTUREALPHA			= D3DTOP_BLENDTEXTUREALPHA,
	GB3DTOP_BLENDFACTORALPHA			= D3DTOP_BLENDFACTORALPHA,
	GB3DTOP_BLENDTEXTUREALPHAPM			= D3DTOP_BLENDTEXTUREALPHAPM,
	GB3DTOP_BLENDCURRENTALPHA			= D3DTOP_BLENDCURRENTALPHA,
	GB3DTOP_PREMODULATE 				= D3DTOP_PREMODULATE,
	GB3DTOP_MODULATEALPHA_ADDCOLOR		= D3DTOP_MODULATEALPHA_ADDCOLOR,
	GB3DTOP_MODULATECOLOR_ADDALPHA		= D3DTOP_MODULATECOLOR_ADDALPHA,
	GB3DTOP_MODULATEINVALPHA_ADDCOLOR	= D3DTOP_MODULATEINVALPHA_ADDCOLOR,
	GB3DTOP_MODULATEINVCOLOR_ADDALPHA	= D3DTOP_MODULATEINVCOLOR_ADDALPHA,
	GB3DTOP_BUMPENVMAP					= D3DTOP_BUMPENVMAP,
	GB3DTOP_BUMPENVMAPLUMINANCE			= D3DTOP_BUMPENVMAPLUMINANCE,
	GB3DTOP_DOTPRODUCT3					= D3DTOP_DOTPRODUCT3
} GB3DTEXTUREOP;

#define GB3DTA_CURRENT			D3DTA_CURRENT
#define GB3DTA_DIFFUSE			D3DTA_DIFFUSE
#define GB3DTA_SELECTMASK		D3DTA_SELECTMASK
#define GB3DTA_TEXTURE			D3DTA_TEXTURE
#define GB3DTA_TFACTOR			D3DTA_TFACTOR
#define GB3DTA_ALPHAREPLICATE	D3DTA_ALPHAREPLICATE
#define GB3DTA_COMPLEMENT		D3DTA_COMPLEMENT

#define GB3DRENDERSTATE_BILINEARFILTER		(1 << 1)
#define GB3DRENDERSTATE_DITHER				(1 << 2)
#define GB3DRENDERSTATE_PERSPECTIVECOLLECT	(1 << 3)

#define GB3DCLEAR_TARGET			D3DCLEAR_TARGET
#define GB3DCLEAR_ZBUFFER			D3DCLEAR_ZBUFFER

typedef DWORD GB3DCOLOR;

typedef struct {
	CHAR	pszName[256];
	CHAR	pszLocalName[256];
	BOOL	bHardware;
} *LPGB3DDEVICEINFO, GB3DDEVICEINFO;

typedef struct {
	CHAR	pszName[256];
	INT		nBits;
} *LPGB3DZBUFFERFORMATINFO, GB3DZBUFFERFORMATINFO;

typedef struct {
	CHAR	pszName[256];
	INT		nBits;
	INT		nRBits;
	INT		nRShifts;
	INT		nRMax;
	INT		nGBits;
	INT		nGShifts;
	INT		nGMax;
	INT		nBBits;
	INT		nBShifts;
	INT		nBMax;
	INT		nABits;
	INT		nAShifts;
	INT		nAMax;
	BOOL	bPalette;
} *LPGB3DTEXTUREFORMATINFO, GB3DTEXTUREFORMATINFO;

typedef struct {
	INT		nRenderTarget;
	INT		nDevice;
	INT		nZBufferFormat;
	DWORD	dwFlags;
} *LPGB3DSTARTUPDEVICEINFO, GB3DSTARTUPDEVICEINFO;

typedef D3DMATERIAL *LPGB3DMATERIALINFO, GB3DMATERIALINFO;

typedef D3DLIGHT2 *LPGB3DLIGHTINFO, GB3DLIGHTINFO;

typedef struct {
	INT   nSrcX;
	INT   nSrcY;
	INT   nSrcW;
	INT   nSrcH;
	INT   nSrcTexture;
	INT   nDstX;
	INT   nDstY;
	DWORD dwFlags;
} *LPGB3DSPRITE, GB3DSPRITE;

typedef struct GB3DTXUV {
	GB3DTXUV() {}
	GB3DTXUV(FLOAT fU, FLOAT fV) { u = fU; v = fV; }
	FLOAT u;
	FLOAT v;
} GB3DTXUV;

typedef struct {
	INT nNumVertices;
	INT nNumVertIdxs;

	GB3DVECTOR* pVert;
	GB3DVECTOR* pNorm;
 	GB3DTXUV*	pTXUV;
	LPWORD pwVertIdx;

	DWORD dwFlags;
	VOID* pModelData;
} *LPGB3DMODEL, GB3DMODEL;

typedef struct tagGB3DCOORDINATE {
	GB3DMATRIX matrix;
	tagGB3DCOORDINATE* pParent;
	BOOL bChanged;
	GB3DMATRIX workmatrix;
} *LPGB3DCOORDINATE, GB3DCOORDINATE;

typedef struct {
	GB3DTEXTUREOP op;
	DWORD dwArg1;
	DWORD dwArg2;
} *LPGB3DTEXTURECOLORSTATE, GB3DTEXTURECOLORSTATE;

typedef struct {
	GB3DTEXTUREOP op;
	DWORD dwArg1;
	DWORD dwArg2;
} *LPGB3DTEXTUREALPHASTATE, GB3DTEXTUREALPHASTATE;

typedef struct {
	DWORD dwAddressU;
	DWORD dwAddressV;
	DWORD dwBorderColor;
} *LPGB3DTEXTURECOORDINATESTATE, GB3DTEXTURECOORDINATESTATE;

typedef struct {
	LPINT pnTexture;
	GB3DTEXTURECOLORSTATE colorstate;
	GB3DTEXTUREALPHASTATE alphastate;
	GB3DTEXTURECOORDINATESTATE coordinatestate;
} *LPGB3DOBJECTTEXTURE, GB3DOBJECTTEXTURE;

typedef struct tagGB3DOBJECT {
	DWORD dwFlags;
	DWORD dwEffects;
	D3DBLEND blendSrc;
	D3DBLEND blendDst;
	D3DTEXTUREBLEND blendmode;
	GB3DOBJECTTEXTURE texture[8];
	GB3DCOORDINATE* pCoord;
	LPINT pnModel;
	LPINT pnMaterial;
	tagGB3DOBJECT* pNext;
} *LPGB3DOBJECT, GB3DOBJECT;

// Initialize

GBEXPORT BOOL GB3DStartup();
GBEXPORT VOID GB3DCleanup();

GBEXPORT VOID GB3DEnableHardwareDevice(BOOL bHAL = TRUE);

GBEXPORT INT  GB3DGetDeviceCount();
GBEXPORT INT  GB3DGetAutoDetectedDevice();
GBEXPORT VOID GB3DGetDeviceInfo(LPGB3DDEVICEINFO pInfo, INT nIndex);

GBEXPORT INT  GB3DGetZBufferFormatCount(INT nDevice = -1);
GBEXPORT INT  GB3DGetAutoDetectedZBufferFormat(INT nDevice = -1);
GBEXPORT VOID GB3DGetZBufferFormatInfo(LPGB3DZBUFFERFORMATINFO pInfo, INT nIndex);

GBEXPORT INT  GB3DGetTextureFormatCount();
GBEXPORT VOID GB3DGetTextureFormatInfo(LPGB3DTEXTUREFORMATINFO pInfo, INT nIndex);

GBEXPORT BOOL GB3DStartupDevice(LPGB3DSTARTUPDEVICEINFO pInfo = NULL);
GBEXPORT VOID GB3DCleanupDevice();

// Splash

GBEXPORT BOOL GB3DSplashDirectX();
GBEXPORT BOOL GB3DSplashLetSoftware();

// State

GBEXPORT VOID GB3DSetProjectionMatrix(GB3DMATRIX* pmProj);
GBEXPORT VOID GB3DSetViewMatrix(GB3DMATRIX* pmView);
GBEXPORT VOID GB3DSetAmbient(GB3DCOLOR cAmbient);

GBEXPORT VOID GB3DShowNormal(BOOL bShow, FLOAT fScale = 16.0f);

GBEXPORT VOID GB3DSetRenderState(DWORD dwType, BOOL bOn);

GBEXPORT BOOL GB3DSetRenderTarget(INT nBuffer);

// Render

GBEXPORT VOID GB3DRender(LPGB3DOBJECT pObj);
GBEXPORT VOID GB3DRenderFast(LPGB3DOBJECT pObj);

GBEXPORT BOOL GB3DBeginRender(INT nViewport = GB3DVIEWPORT_DEFAULT);
GBEXPORT BOOL GB3DEndRender();
GBEXPORT VOID GB3DClear(GB3DCOLOR color = 0, FLOAT fZValue = 1.0f,
						DWORD dwFlags = GB3DCLEAR_TARGET | GB3DCLEAR_ZBUFFER,
						INT nViewport = GB3DVIEWPORT_DEFAULT);

// Render 2

GBEXPORT VOID GB3DClearOrderList();
GBEXPORT VOID GB3DOrderObject(LPGB3DOBJECT pObj);
GBEXPORT VOID GB3DRenderAll();

// Object

GBEXPORT INT  GB3DCreateViewport(FLOAT fLeft = -1.0f, FLOAT fTop = -1.0f, FLOAT fRight = 1.0f, FLOAT fBottom = 1.0f);
GBEXPORT BOOL GB3DSetViewport(INT nViewport, FLOAT fLeft = -1.0f, FLOAT fTop = -1.0f, FLOAT fRight = 1.0f, FLOAT fBottom = 1.0f);
GBEXPORT VOID GB3DReleaseViewport(INT nViewport);
GBEXPORT VOID GB3DReleaseAllViewports();

GBEXPORT INT  GB3DCreateTexture(LPCTSTR pszFile, INT nTextureFormat = -1);
GBEXPORT INT  GB3DCreateTexture(HBITMAP hbitmap, INT nTextureFormat = -1);
GBEXPORT VOID GB3DReleaseTexture(INT nTexture);

GBEXPORT INT  GB3DCreateMaterial(LPGB3DMATERIALINFO pInfo);
GBEXPORT VOID GB3DSetMaterial(INT nMaterial, LPGB3DMATERIALINFO pInfo);
GBEXPORT VOID GB3DReleaseMaterial(INT nMaterial);
GBEXPORT VOID GB3DReleaseAllMaterials();

GBEXPORT INT  GB3DCreateLight(LPGB3DLIGHTINFO pInfo);
GBEXPORT VOID GB3DSetLight(INT nLight, LPGB3DLIGHTINFO pInfo);
GBEXPORT VOID GB3DAddLight(INT nLight, INT nViewport = GB3DVIEWPORT_DEFAULT);
GBEXPORT VOID GB3DDeleteLight(INT nLight, INT nViewport = GB3DVIEWPORT_DEFAULT);
GBEXPORT VOID GB3DReleaseLight(INT nLight);
GBEXPORT VOID GB3DReleaseAllLights();

GBEXPORT INT  GB3DCreateModel(LPCTSTR pszFile, DWORD dwFlags = GB3DMODEL_FLAT);
GBEXPORT INT  GB3DCreateModel(LPVOID pData, DWORD dwFlags = GB3DMODEL_FLAT);
GBEXPORT VOID GB3DBuildModelVertex(INT nModel, DWORD dwFlags = GB3DMODEL_FLAT);
GBEXPORT VOID GB3DBuildModelLVertex(INT nModel);
GBEXPORT VOID GB3DReleaseModel(INT nModel);
GBEXPORT VOID GB3DReleaseAllModels();

GBEXPORT VOID GB3DReleaseAllResources();

GBEXPORT VOID GB3DDrawSprite(
	INT nSrcX, INT nSrcY, INT nSrcW, INT nSrcH, INT nSrcTexture,
	INT nDstX, INT nDstY, INT nDstBuffer = GBVIDEO_BACKBUFFER, DWORD dwFlags = 0);
GBEXPORT VOID GB3DDrawSprite(LPGB3DSPRITE pgb3dsprite, INT nDstBuffer = GBVIDEO_BACKBUFFER);
GBEXPORT VOID GB3DDrawSpriteBatch(LPGB3DSPRITE pgb3dsprite, INT nCount, INT nDstBuffer = GBVIDEO_BACKBUFFER);
GBEXPORT VOID GB3DDrawSpriteFastBatch(LPGB3DSPRITE pgb3dsprite, INT nCount,
	INT nSrcBuffer, INT nDstBuffer = GBVIDEO_BACKBUFFER);

GBEXPORT BOOL GB3DRestoreAllResources();

// DirectX Interface

GBEXPORT VOID GB3DGetDirect3DInterface(LPDIRECT3D3* pd3d);
GBEXPORT VOID GB3DGetDirect3DDeviceInterface(LPDIRECT3DDEVICE3* pd3ddevice);

//VOID GB3DGetD3DTextureHandle(LPD3DTEXTUREHANDLE, INT nTexture);
GBEXPORT VOID GB3DGetDirect3DTextureInterface(LPDIRECT3DTEXTURE2* pd3dt, INT nTexture);
GBEXPORT VOID GB3DGetD3DMaterialHandle(LPD3DMATERIALHANDLE, INT nMaterial);

/////////////////////////////////////////////////////////////////////////////
#endif
/////////////////////////////////////////////////////////////////////////////
