/////////////////////////////////////////////////////////////////////////////
// GliD3D_texture.cpp : No comment
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// Directive

#include "StdHdr.h"
#pragma hdrstop
#include "GliD3D.h"

/////////////////////////////////////////////////////////////////////////////
// Defines / Variables

typedef struct {
	DDPIXELFORMAT	ddpf;
	CHAR			pszName[256];
	INT				nBits;
	INT				nR;
	INT				nG;
	INT				nB;
	INT				nA;
	BOOL			bPalette;
} *LPGB3DCORETEXTUREFORMATINFO, GB3DCORETEXTUREFORMATINFO;

/////////////////////////////////////////////////////////////////////////////
// Prototypes

VOID GB3DCoreGetTextureFormatInfo(LPGB3DCORETEXTUREFORMATINFO pInfo, INT nIndex);

/////////////////////////////////////////////////////////////////////////////
// Implements

FX_ENTRY FxU32 FX_CALL 
grTexCalcMemRequired(
                     GrLOD_t lodmin, GrLOD_t lodmax,
                     GrAspectRatio_t aspect, GrTextureFormat_t fmt)
{
	TRACEFUNCTION("grTexCalcMemRequired");

	return GLID3D_TEXCALCMEMREQUIRED;
}

FX_ENTRY FxU32 FX_CALL 
grTexTextureMemRequired( FxU32     evenOdd,
                                 GrTexInfo *info   )
{
	TRACEFUNCTION("grTexTextureMemRequired");

	return GLID3D_TEXTEXTUREMEMREQUIRED;
}

FX_ENTRY FxU32 FX_CALL 
grTexMinAddress( GrChipID_t tmu )
{
	TRACEFUNCTION("grTexMinAddress");

	return GLID3D_TEXMINADDRESS;
}

FX_ENTRY FxU32 FX_CALL 
grTexMaxAddress( GrChipID_t tmu )
{
	TRACEFUNCTION("grTexMaxAddress");

	return GLID3D_TEXMAXADDRESS;
}

FX_ENTRY void FX_CALL 
grTexNCCTable( GrNCCTable_t table )
{
	TRACEFUNCTION("grTexNCCTable");

	WARNING("Not supported yet");
}

FX_ENTRY void FX_CALL 
grTexSource( GrChipID_t tmu,
             FxU32      startAddress,
             FxU32      evenOdd,
             GrTexInfo  *info )
{
	TRACEFUNCTION("grTexSource");

	if (startAddress < 0 || startAddress >= GLID3D_TEXMAXADDRESS) {
		WARNING("Wrong source");
		return;
	}

	// For PSEmu Pro :-)
	if (startAddress > 4096) startAddress /= 131072;

	INT nTmu = grsubGetTmuIndex(tmu);
	g_gr_tmu[nTmu].texture_source_start_address = startAddress;
	g_gr_tmu[nTmu].texture_source_even_odd = evenOdd;
	g_gr_tmu[nTmu].texture_source_info = *info;
}

FX_ENTRY void FX_CALL 
grTexClampMode(
               GrChipID_t tmu,
               GrTextureClampMode_t s_clampmode,
               GrTextureClampMode_t t_clampmode
               )
{
	TRACEFUNCTION("grTexClampMode");

	INT nTmu = grsubGetTmuIndex(tmu);
	g_gr_tmu[nTmu].s_clampmode = s_clampmode;
	g_gr_tmu[nTmu].t_clampmode = t_clampmode;
	g_gr_pd3ddevice->SetTextureStageState(nTmu, D3DTSS_ADDRESSU, s_clampmode == GR_TEXTURECLAMP_WRAP ? D3DTADDRESS_WRAP : D3DTADDRESS_CLAMP);
	g_gr_pd3ddevice->SetTextureStageState(nTmu, D3DTSS_ADDRESSV, t_clampmode == GR_TEXTURECLAMP_WRAP ? D3DTADDRESS_WRAP : D3DTADDRESS_CLAMP);
}

FX_ENTRY void FX_CALL 
grTexCombine(
             GrChipID_t tmu,
             GrCombineFunction_t rgb_function,
             GrCombineFactor_t rgb_factor, 
             GrCombineFunction_t alpha_function,
             GrCombineFactor_t alpha_factor,
             FxBool rgb_invert,
             FxBool alpha_invert
             )
{
	TRACEFUNCTION("grTexCombine");

	INT nTmu = grsubGetTmuIndex(tmu);
	g_gr_tmu[nTmu].texture_combine_rgb_function = rgb_function;
	g_gr_tmu[nTmu].texture_combine_rgb_factor = rgb_factor;
	g_gr_tmu[nTmu].texture_combine_alpha_function = alpha_function;
	g_gr_tmu[nTmu].texture_combine_alpha_factor = alpha_factor;
}

FX_ENTRY void FX_CALL 
grTexCombineFunction(
                     GrChipID_t tmu,
                     GrTextureCombineFnc_t fnc
                     )
{
	TRACEFUNCTION("grTexCombineFunction");

	switch (fnc) {
	case GR_TEXTURECOMBINE_ZERO          : break;
	case GR_TEXTURECOMBINE_DECAL         : break;
	case GR_TEXTURECOMBINE_OTHER         : break;
	case GR_TEXTURECOMBINE_ADD           : break;
	case GR_TEXTURECOMBINE_MULTIPLY      : break;
	case GR_TEXTURECOMBINE_SUBTRACT      : break;
	case GR_TEXTURECOMBINE_DETAIL        : break;
	case GR_TEXTURECOMBINE_DETAIL_OTHER  : break;
	case GR_TEXTURECOMBINE_TRILINEAR_ODD : break;
	case GR_TEXTURECOMBINE_TRILINEAR_EVEN: break;
	case GR_TEXTURECOMBINE_ONE           : break;
	}
}

FX_ENTRY void FX_CALL 
grTexDetailControl(
                   GrChipID_t tmu,
                   int lod_bias,
                   FxU8 detail_scale,
                   float detail_max
                   )
{
	TRACEFUNCTION("grTexDetailControl");

	WARNING("Not supported yet");
}

FX_ENTRY void FX_CALL 
grTexFilterMode(
                GrChipID_t tmu,
                GrTextureFilterMode_t minfilter_mode,
                GrTextureFilterMode_t magfilter_mode
                )
{
	TRACEFUNCTION("grTexFilterMode");

	INT nTmu = grsubGetTmuIndex(tmu);
	g_gr_tmu[nTmu].texture_minfilter_mode = minfilter_mode;
	g_gr_tmu[nTmu].texture_magfilter_mode = magfilter_mode;
	g_gr_pd3ddevice->SetTextureStageState(nTmu, D3DTSS_MINFILTER,
		minfilter_mode == GR_TEXTUREFILTER_POINT_SAMPLED ? D3DTFN_POINT : D3DTFN_LINEAR);
	g_gr_pd3ddevice->SetTextureStageState(nTmu, D3DTSS_MAGFILTER,
		magfilter_mode == GR_TEXTUREFILTER_POINT_SAMPLED ? D3DTFN_POINT : D3DTFN_LINEAR);
}


FX_ENTRY void FX_CALL 
grTexLodBiasValue(GrChipID_t tmu, float bias )
{
	TRACEFUNCTION("grTexLodBiasValue");

	WARNING("Not supported yet");
}

BOOL
grsubCreateSurface(LPDIRECTDRAWSURFACE4* ppddsurface, LPINT pnTexFormat, GrTexInfo* info)
{
	// Large lod
	INT nLargeLod;
	nLargeLod = grsubLodToInt(info->largeLodLog2);

	// Aspect ratio
	INT nAspectW, nAspectH;
	grsubAspectToInt(&nAspectW, &nAspectH, info->aspectRatioLog2);

	// Width and Height
	INT nWidth, nHeight;
	nWidth = nLargeLod / nAspectW;
	nHeight = nLargeLod / nAspectH;


	// Create device surface
	HRESULT hr;
	LPDIRECTDRAWSURFACE4 pdds;
	pdds = NULL;
	DDSURFACEDESC2 ddsd;
	ZeroMemory(&ddsd, sizeof(ddsd));
	ddsd.dwSize = sizeof(ddsd);
	ddsd.dwFlags = DDSD_CAPS | DDSD_WIDTH | DDSD_HEIGHT | DDSD_PIXELFORMAT;
	ddsd.ddsCaps.dwCaps2 = DDSCAPS2_TEXTUREMANAGE;
	if (((nWidth / g_gr_nReduceTexture) >= 2) && ((nHeight / g_gr_nReduceTexture) >= 2)) {
		ddsd.dwWidth = nWidth / g_gr_nReduceTexture;
		ddsd.dwHeight = nHeight / g_gr_nReduceTexture;
	} else {
		ddsd.dwWidth = nWidth;
		ddsd.dwHeight = nHeight;
	}

	*pnTexFormat = g_gr_texfmt[info->format].nFormat;
	
	static GB3DCORETEXTUREFORMATINFO texinfo;
	GB3DCoreGetTextureFormatInfo(&texinfo, *pnTexFormat);
	ddsd.ddpfPixelFormat = texinfo.ddpf;
	ddsd.ddsCaps.dwCaps = DDSCAPS_TEXTURE;
	hr = g_gr_pdd->CreateSurface(&ddsd, &pdds, NULL);
	if (FAILED(hr)) {
		WARNING("Couldn't create texture surface");
		return FALSE;
	}

	*ppddsurface = pdds;

	return TRUE;
}

BOOL
grsubCreatePalette(LPDIRECTDRAWPALETTE* ppddpalette, LPINT pnTexFormat, GrTexInfo* info)
{
	HRESULT hr;
	LPDIRECTDRAWPALETTE pddp;
	pddp = NULL;
	hr = g_gr_pdd->CreatePalette(DDPCAPS_8BITENTRIES, NULL, &pddp, NULL);
	if (FAILED(hr)) {
		WARNING("Couldn't create texture palette");
		return FALSE;
	}

	*ppddpalette = pddp;

	return TRUE;
}

FX_ENTRY void FX_CALL 
grTexDownloadMipMap( GrChipID_t tmu,
                     FxU32      startAddress,
                     FxU32      evenOdd,
                     GrTexInfo  *info )
{
	TRACEFUNCTION("grTexDownloadMipMap");

	INT nLargeLod = grsubLodToInt(info->largeLodLog2);
	INT nAspectW, nAspectH;
	grsubAspectToInt(&nAspectW, &nAspectH, info->aspectRatioLog2);

	HRESULT hr;
	RECT rcDest;
	rcDest.left = 0;
	rcDest.top = 0;
	rcDest.right = nLargeLod / nAspectW;
	rcDest.bottom = nLargeLod / nAspectH;

	LPDIRECTDRAWSURFACE4 pddsurface = NULL;
	LPDIRECT3DTEXTURE2 pd3dtexture = NULL;
	LPDIRECTDRAWPALETTE pddpalette = NULL;
	INT nTmu = grsubGetTmuIndex(tmu);
	LPGLID3D_TMU ptmu = &g_gr_tmu[nTmu];
	LPGLID3D_TEXTURE ptexture = &ptmu->texture[startAddress];
	INT nTexFormat = 0;
	if (ptexture->pd3dtexture) {
		if (ptexture->nWidth != rcDest.right || ptexture->nHeight != rcDest.bottom || ptexture->info.format != info->format) {
			RELEASE(ptexture->pd3dtexture);
			RELEASE(ptexture->pddsurface);
			ptexture->pddsurface = NULL;
			ptexture->pd3dtexture = NULL;
		} else {

//			return;
			pddsurface = ptexture->pddsurface;
			pd3dtexture = ptexture->pd3dtexture;
			nTexFormat = g_gr_texfmt[info->format].nFormat;

		}
	}

	if (!pd3dtexture) {

		if (grsubCreateSurface(&pddsurface, &nTexFormat, info)) {
			ptexture->pddsurface = pddsurface;
			pddsurface->QueryInterface(IID_IDirect3DTexture2, (LPVOID*)&pd3dtexture);
			if (ptexture->bPal) {
				grsubCreatePalette(&pddpalette, &nTexFormat, info);
				ptexture->pddpalette = pddpalette;
			}
		} else {
			ptexture->pd3dtexture = NULL;
			ptexture->pddsurface = NULL;
			return;
		}
		ptexture->pd3dtexture = pd3dtexture;

	}
	ptexture->info = *info;
	ptexture->nWidth = rcDest.right;
	ptexture->nHeight = rcDest.bottom;

	TRACE("GliD3D info : DownloadMipMap - Address = %d, Size(%d, %d)\n", startAddress, rcDest.right, rcDest.bottom);

	// Lock the surface and Write pixel to it
//	GLID3D_PIXELFORMAT pf;
//	grsubGetSurfacePixelFormat(&pf, pddsurface);

//	TRACE("GliD3D info : Lock the surface\n");
	INT nPitch;
	DDSURFACEDESC2 ddsd;
	ZeroMemory(&ddsd, sizeof(ddsd));
	ddsd.dwSize = sizeof(ddsd);
	hr = pddsurface->GetSurfaceDesc(&ddsd);
	nPitch = ddsd.lPitch;
	hr = pddsurface->Lock(NULL, &ddsd, DDLOCK_WRITEONLY | DDLOCK_WAIT, NULL);
	if (SUCCEEDED(hr)) {
//		TRACE("GliD3D info : Write pixel to Locked surface\n");
		LPBYTE pbyData = (LPBYTE)info->data;
		LPWORD pwData = (LPWORD)info->data;
		LPDWORD pdwData = (LPDWORD)info->data;
		LPBYTE pbySurf = (LPBYTE)ddsd.lpSurface;
		LPWORD pwSurf = (LPWORD)ddsd.lpSurface;
		LPDWORD pdwSurf = (LPDWORD)ddsd.lpSurface;
		DWORD dwSrc, dwDst;

		INT nBits = g_gr_texture_format[nTexFormat].nBits;
		INT nRBits = g_gr_texture_format[nTexFormat].nRBits;
		INT nRShifts = g_gr_texture_format[nTexFormat].nRShifts;
		INT nRMax = g_gr_texture_format[nTexFormat].nRMax;
		INT nGBits = g_gr_texture_format[nTexFormat].nGBits;
		INT nGShifts = g_gr_texture_format[nTexFormat].nGShifts;
		INT nGMax = g_gr_texture_format[nTexFormat].nGMax;
		INT nBBits = g_gr_texture_format[nTexFormat].nBBits;
		INT nBShifts = g_gr_texture_format[nTexFormat].nBShifts;
		INT nBMax = g_gr_texture_format[nTexFormat].nBMax;
		INT nABits = g_gr_texture_format[nTexFormat].nABits;
		INT nAShifts = g_gr_texture_format[nTexFormat].nAShifts;
		INT nAMax = g_gr_texture_format[nTexFormat].nAMax;

		static INT x, y, nXCount, nYCount;
		BOOL bReduce;
		if (rcDest.right / g_gr_nReduceTexture >= 2 && rcDest.bottom / g_gr_nReduceTexture >= 2) {
			nXCount = rcDest.right / g_gr_nReduceTexture;
			nYCount = rcDest.bottom / g_gr_nReduceTexture;
			bReduce = TRUE;
		} else {
			nXCount = rcDest.right;
			nYCount = rcDest.bottom;
			bReduce = FALSE;
		}
		switch (g_gr_state.color_format) {
		case GR_COLORFORMAT_ABGR:
			{
				switch (info->format) {
				case GR_TEXFMT_RGB_332            :
					{
						TRACE("GliD3D info : GR_TEXFMT_RGB_332\n");
						break;
					}
				case GR_TEXFMT_YIQ_422            :
					{
						TRACE("GliD3D info : GR_TEXFMT_YIQ_422\n");
						break;
					}
				case GR_TEXFMT_ALPHA_8            :
					{
						TRACE("GliD3D info : GR_TEXFMT_ALPHA_8\n");
						break;
					}
				case GR_TEXFMT_INTENSITY_8        :
					{
						TRACE("GliD3D info : GR_TEXFMT_INTENSITY_8\n");
						break;
					}
				case GR_TEXFMT_ALPHA_INTENSITY_44 :
					{
						TRACE("GliD3D info : GR_TEXFMT_ALPHA_INTENSITY_44\n");
						break;
					}
				case GR_TEXFMT_P_8                :
					{
						TRACE("GliD3D info : GR_TEXFMT_P_8\n");
						switch (nBits) {
						case 8:
							{
								for (y = 0; y < nYCount; y++) {
									pbySurf = (LPBYTE)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									pwSurf = (LPWORD)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									pdwSurf = (LPDWORD)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									for (x = 0; x < nXCount; x++) {
										dwSrc = (DWORD)*pbyData++;
										dwDst = dwSrc;
										*pbySurf++ = (BYTE)dwDst;
										pbyData += (bReduce ? (g_gr_nReduceTexture - 1) : 0);
									}
									pbyData += (bReduce ? nXCount * g_gr_nReduceTexture * (g_gr_nReduceTexture - 1) : 0);
								}
								break;
							}
						case 16:
							{
								for (y = 0; y < nYCount; y++) {
									pbySurf = (LPBYTE)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									pwSurf = (LPWORD)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									pdwSurf = (LPDWORD)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									for (x = 0; x < nXCount; x++) {
										dwSrc = (DWORD)*pbyData++;
										dwDst = (((dwSrc & 0x0000001fL) >>  0) * (nBMax + 0) /  31 << nBShifts) |
												(((dwSrc & 0x000007e0L) >>  5) * (nGMax + 0) /  63 << nGShifts) |
												(((dwSrc & 0x0000f800L) >> 11) * (nRMax + 0) /  31 << nRShifts);
										if ((dwSrc & 0xffff) != 0 && dwDst == 0) dwDst = 1 << nGShifts;
										*pwSurf++ = (WORD)dwDst;
										pbyData += (bReduce ? (g_gr_nReduceTexture - 1) : 0);
									}
									pbyData += (bReduce ? nXCount * g_gr_nReduceTexture * (g_gr_nReduceTexture - 1) : 0);
								}
								break;
							}
						case 24:
						case 32:
							{
								for (y = 0; y < nYCount; y++) {
									pbySurf = (LPBYTE)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									pwSurf = (LPWORD)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									pdwSurf = (LPDWORD)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									for (x = 0; x < nXCount; x++) {
										dwSrc = (DWORD)*pbyData++;
										dwDst = (((dwSrc & 0x0000001fL) >>  0) * (nBMax + 0) /  31 << nBShifts) |
												(((dwSrc & 0x000007e0L) >>  5) * (nGMax + 0) /  63 << nGShifts) |
												(((dwSrc & 0x0000f800L) >> 11) * (nRMax + 0) /  31 << nRShifts);
										if ((dwSrc & 0xffff) != 0 && dwDst == 0) dwDst = 1 << nGShifts;
										*pdwSurf++ = (DWORD)dwDst;
										pbyData += (bReduce ? (g_gr_nReduceTexture - 1) : 0);
									}
									pbyData += (bReduce ? nXCount * g_gr_nReduceTexture * (g_gr_nReduceTexture - 1) : 0);
								}
								break;
							}
						}
						break;
					}
				case GR_TEXFMT_ARGB_8332          :
					{
						TRACE("GliD3D info : GR_TEXFMT_ARGB_8332\n");
						break;
					}
				case GR_TEXFMT_AYIQ_8422          :
					{
						TRACE("GliD3D info : GR_TEXFMT_AYIQ_8422\n");
						break;
					}
				case GR_TEXFMT_RGB_565            :
					{
//						TRACE("GliD3D info : GR_TEXFMT_RGB_565\n");
						switch (nBits) {
						case 8:
							{
								for (y = 0; y < nYCount; y++) {
									pbySurf = (LPBYTE)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									pwSurf = (LPWORD)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									pdwSurf = (LPDWORD)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									for (x = 0; x < nXCount; x++) {
										dwSrc = (DWORD)*pwData++;
										dwDst = (((dwSrc & 0x0000001fL) >>  0) * (nBMax + 0) /  31 << nBShifts) |
												(((dwSrc & 0x000007e0L) >>  5) * (nGMax + 0) /  63 << nGShifts) |
												(((dwSrc & 0x0000f800L) >> 11) * (nRMax + 0) /  31 << nRShifts);
										if ((dwSrc & 0xffff) != 0 && dwDst == 0) dwDst = 1 << nGShifts;
										*pbySurf++ = (BYTE)dwDst;
										pwData += (bReduce ? (g_gr_nReduceTexture - 1) : 0);
									}
									pwData += (bReduce ? nXCount * g_gr_nReduceTexture * (g_gr_nReduceTexture - 1) : 0);
								}
								break;
							}
						case 16:
							{
								for (y = 0; y < nYCount; y++) {
									pbySurf = (LPBYTE)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									pwSurf = (LPWORD)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									pdwSurf = (LPDWORD)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									for (x = 0; x < nXCount; x++) {
										dwSrc = (DWORD)*pwData++;
										dwDst = (((dwSrc & 0x0000001fL) >>  0) * (nBMax + 0) /  31 << nBShifts) |
												(((dwSrc & 0x000007e0L) >>  5) * (nGMax + 0) /  63 << nGShifts) |
												(((dwSrc & 0x0000f800L) >> 11) * (nRMax + 0) /  31 << nRShifts);
										if ((dwSrc & 0xffff) != 0 && dwDst == 0) dwDst = 1 << nGShifts;
										*pwSurf++ = (WORD)dwDst;
										pwData += (bReduce ? (g_gr_nReduceTexture - 1) : 0);
									}
									pwData += (bReduce ? nXCount * g_gr_nReduceTexture * (g_gr_nReduceTexture - 1) : 0);
								}
								break;
							}
						case 24:
						case 32:
							{
								for (y = 0; y < nYCount; y++) {
									pbySurf = (LPBYTE)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									pwSurf = (LPWORD)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									pdwSurf = (LPDWORD)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									for (x = 0; x < nXCount; x++) {
										dwSrc = (DWORD)*pwData++;
										dwDst = (((dwSrc & 0x0000001fL) >>  0) * (nBMax + 0) /  31 << nBShifts) |
												(((dwSrc & 0x000007e0L) >>  5) * (nGMax + 0) /  63 << nGShifts) |
												(((dwSrc & 0x0000f800L) >> 11) * (nRMax + 0) /  31 << nRShifts);
										if ((dwSrc & 0xffff) != 0 && dwDst == 0) dwDst = 1 << nGShifts;
										*pdwSurf++ = (DWORD)dwDst;
										pwData += (bReduce ? (g_gr_nReduceTexture - 1) : 0);
									}
									pwData += (bReduce ? nXCount * g_gr_nReduceTexture * (g_gr_nReduceTexture - 1) : 0);
								}
								break;
							}
						}
						break;
					}
				case GR_TEXFMT_ARGB_1555          :
					{
//						TRACE("GliD3D info : GR_TEXFMT_ARGB_1555\n");
						switch (nBits) {
						case 8:
							{
								for (y = 0; y < nYCount; y++) {
									pbySurf = (LPBYTE)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									pwSurf = (LPWORD)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									pdwSurf = (LPDWORD)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									for (x = 0; x < nXCount; x++) {
										dwSrc = (DWORD)*pwData++;
										dwDst = (((dwSrc & 0x0000001fL) >>  0) * (nBMax + 0) /  31 << nBShifts) |
												(((dwSrc & 0x000003e0L) >>  5) * (nGMax + 0) /  31 << nGShifts) |
												(((dwSrc & 0x00007c00L) >> 10) * (nRMax + 0) /  31 << nRShifts) |
												(((dwSrc & 0x00008000L) >> 15) * (nAMax + 0) /   1 << nAShifts);
										if ((dwSrc & 0x7fff) != 0 && dwDst == 0) dwDst = 1 << nGShifts;
										*pbySurf++ = (BYTE)dwDst;
										pwData += (bReduce ? (g_gr_nReduceTexture - 1) : 0);
									}
									pwData += (bReduce ? nXCount * g_gr_nReduceTexture * (g_gr_nReduceTexture - 1) : 0);
								}
								break;
							}
						case 16:
							{
								for (y = 0; y < nYCount; y++) {
									pbySurf = (LPBYTE)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									pwSurf = (LPWORD)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									pdwSurf = (LPDWORD)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									for (x = 0; x < nXCount; x++) {
										dwSrc = (DWORD)*pwData++;
										dwDst = (((dwSrc & 0x0000001fL) >>  0) * (nBMax + 0) /  31 << nBShifts) |
												(((dwSrc & 0x000003e0L) >>  5) * (nGMax + 0) /  31 << nGShifts) |
												(((dwSrc & 0x00007c00L) >> 10) * (nRMax + 0) /  31 << nRShifts) |
												(((dwSrc & 0x00008000L) >> 15) * (nAMax + 0) /   1 << nAShifts);
										if ((dwSrc & 0x7fff) != 0 && dwDst == 0) dwDst = 1 << nGShifts;
										*pwSurf++ = (WORD)dwDst;
										pwData += (bReduce ? (g_gr_nReduceTexture - 1) : 0);
									}
									pwData += (bReduce ? nXCount * g_gr_nReduceTexture * (g_gr_nReduceTexture - 1) : 0);
								}
								break;
							}
						case 24:
						case 32:
							{
								for (y = 0; y < nYCount; y++) {
									pbySurf = (LPBYTE)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									pwSurf = (LPWORD)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									pdwSurf = (LPDWORD)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									for (x = 0; x < nXCount; x++) {
										dwSrc = (DWORD)*pwData++;
										dwDst = (((dwSrc & 0x0000001fL) >>  0) * (nBMax + 0) /  31 << nBShifts) |
												(((dwSrc & 0x000003e0L) >>  5) * (nGMax + 0) /  31 << nGShifts) |
												(((dwSrc & 0x00007c00L) >> 10) * (nRMax + 0) /  31 << nRShifts) |
												(((dwSrc & 0x00008000L) >> 15) * (nAMax + 0) /   1 << nAShifts);
										if ((dwSrc & 0x7fff) != 0 && dwDst == 0) dwDst = 1 << nGShifts;
										*pdwSurf++ = (DWORD)dwDst;
										pwData += (bReduce ? (g_gr_nReduceTexture - 1) : 0);
									}
									pwData += (bReduce ? nXCount * g_gr_nReduceTexture * (g_gr_nReduceTexture - 1) : 0);
								}
								break;
							}
						}
						break;
					}
				case GR_TEXFMT_ARGB_4444          :
					{
						TRACE("GliD3D info : GR_TEXFMT_ARGB_4444\n");
/*						for (y = 0; y < nYCount; y++) {
							pwSurf = (LPWORD)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
							for (x = 0; x < nXCount; x++) {
								dwSrc = (DWORD)*pwData++;
								dwSrc = (dwSrc & 0x00ffL) << 8 | (dwSrc & 0xff00L) >> 8;
								dwDst = (((dwSrc & 0x0000000fL) >>  0) << nBShifts) |
										(((dwSrc & 0x000000f0L) >>  4) << nGShifts) |
										(((dwSrc & 0x00000f00L) >>  8) << nRShifts) |
										(((dwSrc & 0x0000f000L) >> 12) << nAShifts);
								*pwSurf++ = (WORD)dwDst;
								pwData += (bReduce ? (g_gr_nReduceTexture - 1) : 0);
							}
							pwData += (bReduce ? nXCount * g_gr_nReduceTexture * (g_gr_nReduceTexture - 1) : 0);
						}
*/						break;
					}
				case GR_TEXFMT_ALPHA_INTENSITY_88 :
					{
						TRACE("GliD3D info : GR_TEXFMT_ALPHA_INTENSITY_88\n");
						break;
					}
				case GR_TEXFMT_AP_88              :
					{
						TRACE("GliD3D info : GR_TEXFMT_AP_88\n");
						break;
					}
				}

				break;
			}

		case GR_COLORFORMAT_ARGB:
			{
				switch (info->format) {
				case GR_TEXFMT_RGB_332            :
					{
//						TRACE("GliD3D info : GR_TEXFMT_RGB_332\n");
						switch (nBits) {
						case 8:
							{
								for (y = 0; y < nYCount; y++) {
									pbySurf = (LPBYTE)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									pwSurf = (LPWORD)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									pdwSurf = (LPDWORD)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									for (x = 0; x < nXCount; x++) {
										dwSrc = (DWORD)*pbyData++;
										dwDst = (((dwSrc & 0x00000003L) >>  0) * (nBMax + 0) /   3 << nBShifts) |
												(((dwSrc & 0x0000001cL) >>  2) * (nGMax + 0) /   7 << nGShifts) |
												(((dwSrc & 0x000000e0L) >>  5) * (nRMax + 0) /   7 << nRShifts);
										if ((dwSrc & 0x00ff) != 0 && dwDst == 0) dwDst = 1 << nGShifts;
										*pbySurf++ = (BYTE)dwDst;
										pbyData += (bReduce ? (g_gr_nReduceTexture - 1) : 0);
									}
									pbyData += (bReduce ? nXCount * g_gr_nReduceTexture * (g_gr_nReduceTexture - 1) : 0);
								}
								break;
							}
						case 16:
							{
								for (y = 0; y < nYCount; y++) {
									pbySurf = (LPBYTE)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									pwSurf = (LPWORD)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									pdwSurf = (LPDWORD)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									for (x = 0; x < nXCount; x++) {
										dwSrc = (DWORD)*pbyData++;
										dwDst = (((dwSrc & 0x00000003L) >>  0) * (nBMax + 0) /   3 << nBShifts) |
												(((dwSrc & 0x0000001cL) >>  2) * (nGMax + 0) /   7 << nGShifts) |
												(((dwSrc & 0x000000e0L) >>  5) * (nRMax + 0) /   7 << nRShifts);
										if ((dwSrc & 0x00ff) != 0 && dwDst == 0) dwDst = 1 << nGShifts;
										*pwSurf++ = (WORD)dwDst;
										pbyData += (bReduce ? (g_gr_nReduceTexture - 1) : 0);
									}
									pbyData += (bReduce ? nXCount * g_gr_nReduceTexture * (g_gr_nReduceTexture - 1) : 0);
								}
								break;
							}
						case 24:
						case 32:
							{
								for (y = 0; y < nYCount; y++) {
									pbySurf = (LPBYTE)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									pwSurf = (LPWORD)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									pdwSurf = (LPDWORD)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									for (x = 0; x < nXCount; x++) {
										dwSrc = (DWORD)*pbyData++;
										dwDst = (((dwSrc & 0x00000003L) >>  0) * (nBMax + 0) /   3 << nBShifts) |
												(((dwSrc & 0x0000001cL) >>  2) * (nGMax + 0) /   7 << nGShifts) |
												(((dwSrc & 0x000000e0L) >>  5) * (nRMax + 0) /   7 << nRShifts);
										if ((dwSrc & 0x00ff) != 0 && dwDst == 0) dwDst = 1 << nGShifts;
										*pdwSurf++ = (DWORD)dwDst;
										pbyData += (bReduce ? (g_gr_nReduceTexture - 1) : 0);
									}
									pbyData += (bReduce ? nXCount * g_gr_nReduceTexture * (g_gr_nReduceTexture - 1) : 0);
								}
								break;
							}
						}
						break;
					}
				case GR_TEXFMT_YIQ_422            :
					{
						TRACE("GliD3D info : GR_TEXFMT_YIQ_422\n");
						break;
					}
				case GR_TEXFMT_ALPHA_8            :
					{
//						TRACE("GliD3D info : GR_TEXFMT_ALPHA_8\n");
						switch (nBits) {
						case 8:
							{
								for (y = 0; y < nYCount; y++) {
									pbySurf = (LPBYTE)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									pwSurf = (LPWORD)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									pdwSurf = (LPDWORD)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									for (x = 0; x < nXCount; x++) {
										dwSrc = (DWORD)*pbyData++;
										dwDst = (((dwSrc & 0x000000ffL) >>  0) * (nAMax + 0) / 255 << nAShifts);
										*pbySurf++ = (BYTE)dwDst;
										pbyData += (bReduce ? (g_gr_nReduceTexture - 1) : 0);
									}
									pbyData += (bReduce ? nXCount * g_gr_nReduceTexture * (g_gr_nReduceTexture - 1) : 0);
								}
								break;
							}
						case 16:
							{
								for (y = 0; y < nYCount; y++) {
									pbySurf = (LPBYTE)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									pwSurf = (LPWORD)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									pdwSurf = (LPDWORD)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									for (x = 0; x < nXCount; x++) {
										dwSrc = (DWORD)*pbyData++;
										dwDst = (((dwSrc & 0x000000ffL) >>  0) * (nAMax + 0) / 255 << nAShifts);
										*pwSurf++ = (WORD)dwDst;
										pwData += (bReduce ? (g_gr_nReduceTexture - 1) : 0);
									}
									pbyData += (bReduce ? nXCount * g_gr_nReduceTexture * (g_gr_nReduceTexture - 1) : 0);
								}
								break;
							}
						case 24:
						case 32:
							{
								for (y = 0; y < nYCount; y++) {
									pbySurf = (LPBYTE)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									pwSurf = (LPWORD)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									pdwSurf = (LPDWORD)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									for (x = 0; x < nXCount; x++) {
										dwSrc = (DWORD)*pbyData++;
										dwDst = (((dwSrc & 0x000000ffL) >>  0) * (nAMax + 0) / 255 << nAShifts);
										*pdwSurf++ = (DWORD)dwDst;
										pbyData += (bReduce ? (g_gr_nReduceTexture - 1) : 0);
									}
									pbyData += (bReduce ? nXCount * g_gr_nReduceTexture * (g_gr_nReduceTexture - 1) : 0);
								}
								break;
							}
						}
						break;
					}
				case GR_TEXFMT_INTENSITY_8        :
					{
						TRACE("GliD3D info : GR_TEXFMT_INTENSITY_8\n");
						break;
					}
				case GR_TEXFMT_ALPHA_INTENSITY_44 :
					{
						TRACE("GliD3D info : GR_TEXFMT_ALPHA_INTENSITY_44\n");
						break;
					}
				case GR_TEXFMT_P_8                :
					{
//						TRACE("GliD3D info : GR_TEXFMT_P_8\n");
/*						switch (nBits) {
						case 8:
							{
								for (y = 0; y < nYCount; y++) {
									pbySurf = (LPBYTE)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									pwSurf = (LPWORD)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									pdwSurf = (LPDWORD)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									for (x = 0; x < nXCount; x++) {
										dwSrc = (DWORD)*pbyData++;
										dwDst = dwSrc;
										*pbySurf++ = (BYTE)dwDst;
										pbyData += (bReduce ? (g_gr_nReduceTexture - 1) : 0);
									}
									pbyData += (bReduce ? nXCount * g_gr_nReduceTexture * (g_gr_nReduceTexture - 1) : 0);
								}
								break;
							}
						case 16:
							{
								for (y = 0; y < nYCount; y++) {
									pbySurf = (LPBYTE)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									pwSurf = (LPWORD)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									pdwSurf = (LPDWORD)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									for (x = 0; x < nXCount; x++) {
										dwSrc = (DWORD)*pbyData++;
										dwDst = (((dwSrc & 0x0000001fL) >>  0) * (nBMax + 0) /  31 << nBShifts) |
												(((dwSrc & 0x000007e0L) >>  5) * (nGMax + 0) /  63 << nGShifts) |
												(((dwSrc & 0x0000f800L) >> 11) * (nRMax + 0) /  31 << nRShifts);
										if ((dwSrc & 0xffff) != 0 && dwDst == 0) dwDst = 1 << nGShifts;
										*pwSurf++ = (WORD)dwDst;
										pbyData += (bReduce ? (g_gr_nReduceTexture - 1) : 0);
									}
									pbyData += (bReduce ? nXCount * g_gr_nReduceTexture * (g_gr_nReduceTexture - 1) : 0);
								}
								break;
							}
						}
						break;
*/					}
				case GR_TEXFMT_ARGB_8332          :
					{
//						TRACE("GliD3D info : GR_TEXFMT_ARGB_8332\n");
						switch (nBits) {
						case 8:
							{
								for (y = 0; y < nYCount; y++) {
									pbySurf = (LPBYTE)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									pwSurf = (LPWORD)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									pdwSurf = (LPDWORD)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									for (x = 0; x < nXCount; x++) {
										dwSrc = (DWORD)*pbyData++;
										dwDst = (((dwSrc & 0x00000003L) >>  0) * (nBMax + 0) /   3 << nBShifts) |
												(((dwSrc & 0x0000001cL) >>  2) * (nGMax + 0) /   7 << nGShifts) |
												(((dwSrc & 0x000000e0L) >>  5) * (nRMax + 0) /   7 << nRShifts) |
												(((dwSrc & 0x0000ff00L) >>  8) * (nAMax + 0) / 255 << nAShifts);
										if ((dwSrc & 0x00ff) != 0 && dwDst == 0) dwDst = 1 << nGShifts;
										*pbySurf++ = (BYTE)dwDst;
										pbyData += (bReduce ? (g_gr_nReduceTexture - 1) : 0);
									}
									pbyData += (bReduce ? nXCount * g_gr_nReduceTexture * (g_gr_nReduceTexture - 1) : 0);
								}
								break;
							}
						case 16:
							{
								for (y = 0; y < nYCount; y++) {
									pbySurf = (LPBYTE)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									pwSurf = (LPWORD)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									pdwSurf = (LPDWORD)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									for (x = 0; x < nXCount; x++) {
										dwSrc = (DWORD)*pwData++;
										dwDst = (((dwSrc & 0x00000003L) >>  0) * (nBMax + 0) /   3 << nBShifts) |
												(((dwSrc & 0x0000001cL) >>  2) * (nGMax + 0) /   7 << nGShifts) |
												(((dwSrc & 0x000000e0L) >>  5) * (nRMax + 0) /   7 << nRShifts) |
												(((dwSrc & 0x0000ff00L) >>  8) * (nAMax + 0) / 255 << nAShifts);
										if ((dwSrc & 0x00ff) != 0 && dwDst == 0) dwDst = 1 << nGShifts;
										*pwSurf++ = (WORD)dwDst;
										pwData += (bReduce ? (g_gr_nReduceTexture - 1) : 0);
									}
									pwData += (bReduce ? nXCount * g_gr_nReduceTexture * (g_gr_nReduceTexture - 1) : 0);
								}
								break;
							}
						case 24:
						case 32:
							{
								for (y = 0; y < nYCount; y++) {
									pbySurf = (LPBYTE)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									pwSurf = (LPWORD)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									pdwSurf = (LPDWORD)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									for (x = 0; x < nXCount; x++) {
										dwSrc = (DWORD)*pwData++;
										dwDst = (((dwSrc & 0x00000003L) >>  0) * (nBMax + 0) /   3 << nBShifts) |
												(((dwSrc & 0x0000001cL) >>  2) * (nGMax + 0) /   7 << nGShifts) |
												(((dwSrc & 0x000000e0L) >>  5) * (nRMax + 0) /   7 << nRShifts) |
												(((dwSrc & 0x0000ff00L) >>  8) * (nAMax + 0) / 255 << nAShifts);
										if ((dwSrc & 0x00ff) != 0 && dwDst == 0) dwDst = 1 << nGShifts;
										*pdwSurf++ = (DWORD)dwDst;
										pwData += (bReduce ? (g_gr_nReduceTexture - 1) : 0);
									}
									pwData += (bReduce ? nXCount * g_gr_nReduceTexture * (g_gr_nReduceTexture - 1) : 0);
								}
								break;
							}
						}
						break;
					}
				case GR_TEXFMT_AYIQ_8422          :
					{
						TRACE("GliD3D info : GR_TEXFMT_AYIQ_8422\n");
						break;
					}
				case GR_TEXFMT_RGB_565            :
					{
//						TRACE("GliD3D info : GR_TEXFMT_RGB_565\n");
						switch (nBits) {
						case 8:
							{
								for (y = 0; y < nYCount; y++) {
									pbySurf = (LPBYTE)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									pwSurf = (LPWORD)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									pdwSurf = (LPDWORD)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									for (x = 0; x < nXCount; x++) {
										dwSrc = (DWORD)*pwData++;
										dwDst = (((dwSrc & 0x0000001fL) >>  0) * (nBMax + 0) /  31 << nBShifts) |
												(((dwSrc & 0x000007e0L) >>  5) * (nGMax + 0) /  63 << nGShifts) |
												(((dwSrc & 0x0000f800L) >> 11) * (nRMax + 0) /  31 << nRShifts);
										if ((dwSrc & 0xffff) != 0 && dwDst == 0) dwDst = 1 << nGShifts;
										*pbySurf++ = (BYTE)dwDst;
										pwData += (bReduce ? (g_gr_nReduceTexture - 1) : 0);
									}
									pwData += (bReduce ? nXCount * g_gr_nReduceTexture * (g_gr_nReduceTexture - 1) : 0);
								}
								break;
							}
						case 16:
							{
								for (y = 0; y < nYCount; y++) {
									pbySurf = (LPBYTE)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									pwSurf = (LPWORD)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									pdwSurf = (LPDWORD)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									for (x = 0; x < nXCount; x++) {
										dwSrc = (DWORD)*pwData++;
										dwDst = (((dwSrc & 0x0000001fL) >>  0) * (nBMax + 0) /  31 << nBShifts) |
												(((dwSrc & 0x000007e0L) >>  5) * (nGMax + 0) /  63 << nGShifts) |
												(((dwSrc & 0x0000f800L) >> 11) * (nRMax + 0) /  31 << nRShifts);
										if ((dwSrc & 0xffff) != 0 && dwDst == 0) dwDst = 1 << nGShifts;
										*pwSurf++ = (WORD)dwDst;
										pwData += (bReduce ? (g_gr_nReduceTexture - 1) : 0);
									}
									pwData += (bReduce ? nXCount * g_gr_nReduceTexture * (g_gr_nReduceTexture - 1) : 0);
								}
								break;
							}
						case 24:
						case 32:
							{
								for (y = 0; y < nYCount; y++) {
									pbySurf = (LPBYTE)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									pwSurf = (LPWORD)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									pdwSurf = (LPDWORD)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									for (x = 0; x < nXCount; x++) {
										dwSrc = (DWORD)*pwData++;
										dwDst = (((dwSrc & 0x0000001fL) >>  0) * (nBMax + 0) /  31 << nBShifts) |
												(((dwSrc & 0x000007e0L) >>  5) * (nGMax + 0) /  63 << nGShifts) |
												(((dwSrc & 0x0000f800L) >> 11) * (nRMax + 0) /  31 << nRShifts);
										if ((dwSrc & 0xffff) != 0 && dwDst == 0) dwDst = 1 << nGShifts;
										*pdwSurf++ = (DWORD)dwDst;
										pwData += (bReduce ? (g_gr_nReduceTexture - 1) : 0);
									}
									pwData += (bReduce ? nXCount * g_gr_nReduceTexture * (g_gr_nReduceTexture - 1) : 0);
								}
								break;
							}
						}
						break;
					}
				case GR_TEXFMT_ARGB_1555          :
					{
//						TRACE("GliD3D info : GR_TEXFMT_ARGB_1555\n");
						switch (nBits) {
						case 8:
							{
								for (y = 0; y < nYCount; y++) {
									pbySurf = (LPBYTE)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									pwSurf = (LPWORD)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									pdwSurf = (LPDWORD)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									for (x = 0; x < nXCount; x++) {
										dwSrc = (DWORD)*pwData++;
										dwDst = (((dwSrc & 0x0000001fL) >>  0) * (nBMax + 0) /  31 << nBShifts) |
												(((dwSrc & 0x000003e0L) >>  5) * (nGMax + 0) /  31 << nGShifts) |
												(((dwSrc & 0x00007c00L) >> 10) * (nRMax + 0) /  31 << nRShifts) |
												(((dwSrc & 0x00008000L) >> 15) * (nAMax + 0) /   1 << nAShifts);
										if ((dwSrc & 0x7fff) != 0 && dwDst == 0) dwDst = 1 << nGShifts;
										*pbySurf++ = (BYTE)dwDst;
										pwData += (bReduce ? (g_gr_nReduceTexture - 1) : 0);
									}
									pwData += (bReduce ? nXCount * g_gr_nReduceTexture * (g_gr_nReduceTexture - 1) : 0);
								}
								break;
							}
						case 16:
							{
								for (y = 0; y < nYCount; y++) {
									pbySurf = (LPBYTE)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									pwSurf = (LPWORD)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									pdwSurf = (LPDWORD)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									for (x = 0; x < nXCount; x++) {
										dwSrc = (DWORD)*pwData++;
										dwDst = (((dwSrc & 0x0000001fL) >>  0) * (nBMax + 0) /  31 << nBShifts) |
												(((dwSrc & 0x000003e0L) >>  5) * (nGMax + 0) /  31 << nGShifts) |
												(((dwSrc & 0x00007c00L) >> 10) * (nRMax + 0) /  31 << nRShifts) |
												(((dwSrc & 0x00008000L) >> 15) * (nAMax + 0) /   1 << nAShifts);
										if ((dwSrc & 0x7fff) != 0 && dwDst == 0) dwDst = 1 << nGShifts;
										*pwSurf++ = (WORD)dwDst;
										pwData += (bReduce ? (g_gr_nReduceTexture - 1) : 0);
									}
									pwData += (bReduce ? nXCount * g_gr_nReduceTexture * (g_gr_nReduceTexture - 1) : 0);
								}
								break;
							}
						case 24:
						case 32:
							{
								for (y = 0; y < nYCount; y++) {
									pbySurf = (LPBYTE)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									pwSurf = (LPWORD)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									pdwSurf = (LPDWORD)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									for (x = 0; x < nXCount; x++) {
										dwSrc = (DWORD)*pwData++;
										dwDst = (((dwSrc & 0x0000001fL) >>  0) * (nBMax + 0) /  31 << nBShifts) |
												(((dwSrc & 0x000003e0L) >>  5) * (nGMax + 0) /  31 << nGShifts) |
												(((dwSrc & 0x00007c00L) >> 10) * (nRMax + 0) /  31 << nRShifts) |
												(((dwSrc & 0x00008000L) >> 15) * (nAMax + 0) /   1 << nAShifts);
										if ((dwSrc & 0x7fff) != 0 && dwDst == 0) dwDst = 1 << nGShifts;
										*pdwSurf++ = (DWORD)dwDst;
										pwData += (bReduce ? (g_gr_nReduceTexture - 1) : 0);
									}
									pwData += (bReduce ? nXCount * g_gr_nReduceTexture * (g_gr_nReduceTexture - 1) : 0);
								}
								break;
							}
						}
						break;
					}
				case GR_TEXFMT_ARGB_4444          :
					{
//						TRACE("GliD3D info : GR_TEXFMT_ARGB_4444\n");
						switch (nBits) {
						case 8:
							{
								for (y = 0; y < nYCount; y++) {
									pbySurf = (LPBYTE)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									pwSurf = (LPWORD)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									pdwSurf = (LPDWORD)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									for (x = 0; x < nXCount; x++) {
										dwSrc = (DWORD)*pwData++;
										dwDst = (((dwSrc & 0x0000000fL) >>  0) * (nBMax + 0) /  15 << nBShifts) |
												(((dwSrc & 0x000000f0L) >>  4) * (nGMax + 0) /  15 << nGShifts) |
												(((dwSrc & 0x00000f00L) >>  8) * (nRMax + 0) /  15 << nRShifts) |
												(((dwSrc & 0x0000f000L) >> 12) * (nAMax + 0) /  15 << nAShifts);
										if ((dwSrc & 0x0fff) != 0 && dwDst == 0) dwDst = 1 << nGShifts;
										*pbySurf++ = (BYTE)dwDst;
										pwData += (bReduce ? (g_gr_nReduceTexture - 1) : 0);
									}
									pwData += (bReduce ? nXCount * g_gr_nReduceTexture * (g_gr_nReduceTexture - 1) : 0);
								}
								break;
							}
						case 16:
							{
								for (y = 0; y < nYCount; y++) {
									pbySurf = (LPBYTE)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									pwSurf = (LPWORD)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									pdwSurf = (LPDWORD)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									for (x = 0; x < nXCount; x++) {
										dwSrc = (DWORD)*pwData++;
										dwDst = (((dwSrc & 0x0000000fL) >>  0) * (nBMax + 0) /  15 << nBShifts) |
												(((dwSrc & 0x000000f0L) >>  4) * (nGMax + 0) /  15 << nGShifts) |
												(((dwSrc & 0x00000f00L) >>  8) * (nRMax + 0) /  15 << nRShifts) |
												(((dwSrc & 0x0000f000L) >> 12) * (nAMax + 0) /  15 << nAShifts);
										if ((dwSrc & 0x0fff) != 0 && dwDst == 0) dwDst = 1 << nGShifts;
										*pwSurf++ = (WORD)dwDst;
										pwData += (bReduce ? (g_gr_nReduceTexture - 1) : 0);
									}
									pwData += (bReduce ? nXCount * g_gr_nReduceTexture * (g_gr_nReduceTexture - 1) : 0);
								}
								break;
							}
						case 24:
						case 32:
							{
								for (y = 0; y < nYCount; y++) {
									pbySurf = (LPBYTE)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									pwSurf = (LPWORD)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									pdwSurf = (LPDWORD)((LPBYTE)(ddsd.lpSurface) + y * ddsd.lPitch);
									for (x = 0; x < nXCount; x++) {
										dwSrc = (DWORD)*pwData++;
										dwDst = (((dwSrc & 0x0000000fL) >>  0) * (nBMax + 0) /  15 << nBShifts) |
												(((dwSrc & 0x000000f0L) >>  4) * (nGMax + 0) /  15 << nGShifts) |
												(((dwSrc & 0x00000f00L) >>  8) * (nRMax + 0) /  15 << nRShifts) |
												(((dwSrc & 0x0000f000L) >> 12) * (nAMax + 0) /  15 << nAShifts);
										if ((dwSrc & 0x0fff) != 0 && dwDst == 0) dwDst = 1 << nGShifts;
										*pdwSurf++ = (DWORD)dwDst;
										pwData += (bReduce ? (g_gr_nReduceTexture - 1) : 0);
									}
									pwData += (bReduce ? nXCount * g_gr_nReduceTexture * (g_gr_nReduceTexture - 1) : 0);
								}
								break;
							}
						}
						break;
					}
				case GR_TEXFMT_ALPHA_INTENSITY_88 :
					{
						TRACE("GliD3D info : GR_TEXFMT_ALPHA_INTENSITY_88\n");
						break;
					}
				case GR_TEXFMT_AP_88              :
					{
						TRACE("GliD3D info : GR_TEXFMT_AP_88\n");
						break;
					}
				}

				break;
			}

		}

		// Unlock the surface
		pddsurface->Unlock(NULL);
//		TRACE("GliD3D info : Unlock the surface\n");

	} else {

		TRACE("GliD3D info : Failed to Lock surface\n");
		TRACE("%s\n", GetDirectXError(hr));

	}

	ptmu->texture_source_start_address_old = -1;

}

FX_ENTRY void FX_CALL 
grTexDownloadMipMapLevel( GrChipID_t        tmu,
                          FxU32             startAddress,
                          GrLOD_t           thisLod,
                          GrLOD_t           largeLod,
                          GrAspectRatio_t   aspectRatio,
                          GrTextureFormat_t format,
                          FxU32             evenOdd,
                          void              *data )
{
	TRACEFUNCTION("grTexDownloadMipMapLevel");

	TRACE("StartAddress %d\n", startAddress);
	if (startAddress > 4096) startAddress /= 131072;

	GrTexInfo info;
	info.smallLodLog2 = thisLod;
	info.largeLodLog2 = largeLod;
	info.aspectRatioLog2 = aspectRatio;
	info.format = format;
	info.data = data;
	grTexDownloadMipMap(tmu, startAddress, evenOdd, &info);
}

FX_ENTRY FxBool FX_CALL 
grTexDownloadMipMapLevelPartial( GrChipID_t        tmu,
                                 FxU32             startAddress,
                                 GrLOD_t           thisLod,
                                 GrLOD_t           largeLod,
                                 GrAspectRatio_t   aspectRatio,
                                 GrTextureFormat_t format,
                                 FxU32             evenOdd,
                                 void              *data,
                                 int               start,
                                 int               end )
{
	TRACEFUNCTION("grTexDownloadMipMapLevelPartial");

	GrTexInfo info;
	info.smallLodLog2 = thisLod;
	info.largeLodLog2 = largeLod;
	info.aspectRatioLog2 = aspectRatio;
	info.format = format;
	info.data = data;
	grTexDownloadMipMap(tmu, startAddress, evenOdd, &info);

	return FXTRUE;
}

FX_ENTRY void FX_CALL
grTexDownloadTable( GrTexTable_t type, 
                    void         *data )

{
	TRACEFUNCTION("grTexDownloadTable");

	WARNING("Not supported yet");
}

FX_ENTRY void FX_CALL
grTexDownloadTablePartial( GrTexTable_t type, 
                           void         *data,
                           int          start,
                           int          end )

{
	TRACEFUNCTION("grTexDownloadTablePartial");

	WARNING("Not supported yet");
}

FX_ENTRY void FX_CALL 
grTexMipMapMode( GrChipID_t     tmu, 
                 GrMipMapMode_t mode,
                 FxBool         lodBlend )
{
	TRACEFUNCTION("grTexMipMapMode");

	INT nTmu = grsubGetTmuIndex(tmu);
	g_gr_tmu[nTmu].mipmap_mode = mode;
	g_gr_pd3ddevice->SetTextureStageState(nTmu, D3DTSS_MIPFILTER,
		mode == GR_MIPMAP_DISABLE ? D3DTFN_POINT : D3DTFP_LINEAR);
}

FX_ENTRY void FX_CALL 
grTexMultibase( GrChipID_t tmu,
                FxBool     enable )
{
	TRACEFUNCTION("grTexMultibase");

	WARNING("Not supported yet");
}

FX_ENTRY void FX_CALL
grTexMultibaseAddress( GrChipID_t       tmu,
                       GrTexBaseRange_t range,
                       FxU32            startAddress,
                       FxU32            evenOdd,
                       GrTexInfo        *info )
{
	TRACEFUNCTION("grTexMultibaseAddress");

	WARNING("Not supported yet");
}

/////////////////////////////////////////////////////////////////////////////
